
/*
 * rshaperctl.c - user access to rshaper
 *
 * Copyright (c) 1998,1999 Alessandro Rubini <rubini@prosa.it>
 *
 * This simple module intercepts netif_rx to lower host bandwidth
 * Tested with two ne2000 devices, see the accompanying 8390.patch
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <asm/types.h>
#include <sys/types.h>
#include <regex.h>

#include "rshaper.h"

#define DEFAULT_TIMEOUT 5000 /* milliseconds */

/*
 * A function to decode host names/numbers
 */
__u32 host_from_string(char *string)
{
    __u32 ip;
    struct hostent *host;

    ip = inet_addr(string);
    if (ip == (__u32)-1) {
        host = gethostbyname(string);
        if (host) ip = *(__u32*)(host->h_addr);
        if (ip == (__u32)-1) ip = 0; /* error */
    }
    return ip;
}

/*
 * A function to decode masks: accept both "/255.255.255.0" and "/24"
 */
__u32 mask_from_string(char *string)
{
    __u32 ip; int i;
    static char trash[4];

    /* check for simple "/24" form */
    if (strlen(string)<5 && isdigit(string[0])
        && sscanf(string,"%i%s",&i,trash)==1) {
        if (i>32) return 0; /* error */
        return htonl(~( (1 << (32-i)) - 1));
    }
    /* otherwise, dotted form */
    ip = inet_addr(string);
    if (ip == (__u32)-1) {
        /* FIXME: use gethostbyname() */
        return 0; /* error */
    }
    return ip;
}

/*
 *  * A function to decode ports
 *   */
__u16 port_from_string(char *string)
{
    int i;

    if (sscanf(string, "%d", &i) == 1)
        if (i > 0 && i < 65536)
	    return i;

    return 0; /* error */
}


volatile void help(char *prgname)
{
    fprintf(stderr, "%s: Usage: \"%s [options] "
	    "[<host>[/<netmask>][:<range>] <bytes-per-sec> [<queue-len>]]\n"
	    "  <host> can specify an host like \"zaigor\" or \"192.168.32.1\"\n"
	    "  <netmask> can specify a netmask like \"/24\" or \"/255.255.255.0\"\n"
	    "  <range> can specify a port range in the form \"<port>-<port>\"\n"
            "  valid options for printout are:\n"
            "     \"-n\" (numeric printout)\n"
            "     \"-t\" (no title line on printout)\n"
	    "see the man page for more information\n",
	    prgname, prgname);
    exit(1);
}

int opt_num = 0;
int opt_title = 1;

int main(int argc, char **argv)
{
    struct rshaper_userinfo info;
    char *ifname = "rshaper";
    char *prgname = argv[0];
    struct ifreq req;
    char sillystring[20];
    regex_t preg;
    regmatch_t pmatch[9];
    int i, sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_IP);

    if (sock < 0) {
        fprintf(stderr, "%s: socket(): %s\n", prgname, strerror(errno));
        exit(1);
    }

    strcpy(req.ifr_name, ifname);
    req.ifr_data = (caddr_t)&info;
    
    /* simplest option parser */
    while (argc > 1 && argv[1][0]=='-') {
        for (i=1; i<strlen(argv[1]); i++) {
            switch(argv[1][i]) {
                case 'n':
                    opt_num = 1; break;
                case 't':
                    opt_title = 0; break;
                default:
                    help(prgname);
            }
        }
        argc--; argv++;
    }

    switch(argc) {

      case 3: /* set the bandwidth for an host, timeout is default */
        info.timeout = DEFAULT_TIMEOUT;
        /* fall through */

      case 4: /* set the bandwidth with speficied timeout */

        if (argc == 4) {
	    float a;
            if (strlen(argv[3]) > 10)
                help(prgname);
            if (sscanf(argv[3],"%f%s", &a, sillystring) != 1)
                help(prgname);
	    info.timeout = (int)(a * 1000); /* milliseconds */
        }

	if (regcomp(&preg, "([a-z0-9_.-]*)?(/([0-9.]*))?(:(([0-9]*)(-([0-9]*))?))?", REG_EXTENDED|REG_ICASE)) {
	    fprintf(stderr, "%s: regcomp(): invalid regexp\n", prgname);
	    exit(1);
	}
	if (regexec(&preg, argv[1], 9, pmatch, 0) == REG_NOMATCH) help(prgname);

	/* check for IP addr */
	if (pmatch[1].rm_so == -1)
	    info.ip = 0;
	else {
	    argv[1][pmatch[1].rm_eo] = '\0';
	    info.ip = host_from_string(&argv[1][pmatch[1].rm_so]);
	}

	/* check for mask */
	if (pmatch[3].rm_so == -1)
	     info.mask = 0xffffffff; /* default: one host */
	else {
	    argv[1][pmatch[3].rm_eo] = '\0';
	    info.mask = mask_from_string(&argv[1][pmatch[3].rm_so]);
	}

	/* check for min port */
	if (pmatch[6].rm_so == -1)
	    info.portmin = 0;
	else
	    if (pmatch[6].rm_so == pmatch[6].rm_eo)
	        info.portmin = 1;
	    else
	        info.portmin = port_from_string(&argv[1][pmatch[6].rm_so]);

	/* check for max port */
	if (pmatch[8].rm_so == -1)
	    info.portmax = info.portmin;
	else {
	    if (pmatch[8].rm_so == pmatch[8].rm_eo)
	        info.portmax = 65535;
	    else
	        info.portmax = port_from_string(&argv[1][pmatch[8].rm_so]);
	}
	if (info.ip == 0 && info.portmin == 0 && info.portmax == 0) help(prgname);
	if (info.portmax != 0 && info.portmin > info.portmax) {
	    fprintf(stderr, "%s: warning! Invalid ports range\n", prgname);
	    exit(1);
	}

	if (strlen(argv[2]) > 10) help(prgname);
	if (sscanf(argv[2],"%d%s", &info.bytes_per_second, sillystring) != 1)
	    help(prgname);

	if (ioctl(sock, SIOCRSHAPERSETINFO, &req)<0) {
	    fprintf(stderr, "%s: ioctl(): %s\n", prgname, strerror(errno));
	    exit(1);
	}
	return 0;

      case 1:
        /*
         * print report of current configuration
         */
        if (opt_title) 
            printf("    host                 mask                 ports       "
                   "bytes-per-sec      max-queue-len (s)\n");

	for (i=0; ; i++) {
	    struct in_addr ip, mask;
            char ip_a[20], mask_a[20]; /* inet_ntoa returns a static ptr... */
            struct hostent *host;
	    info.ip = i;
	    if (ioctl(sock, SIOCRSHAPERGETINFO, &req)<0) {
		if (errno == ENODEV) break; /* ok, no more items */
		fprintf(stderr, "%s: ioctl(): %s\n", prgname, strerror(errno));
		exit(1);
	    }
	    ip.s_addr   = info.ip;   strcpy(ip_a,   inet_ntoa(ip));
	    mask.s_addr = info.mask; strcpy(mask_a, inet_ntoa(mask));
	    host = NULL;
	    if (!strcmp(mask_a, "255.255.255.255"))
		host = gethostbyaddr((char *)&ip, 4, AF_INET);
            
            printf ("%-20s %-20s  %5d %5d      %6i               %7.3f\n",
                    (opt_num || !host) ? ip_a : host->h_name, mask_a,
		    info.portmin, info.portmax,
                    info.bytes_per_second, (float)info.timeout/1000);
	}
	return 0;
	
      default:
	     help(prgname);
    }
    return 0;
}
